"""Integration with OpenAI's API."""
import functools
import os
from typing import Callable, Dict, List, Optional, Union

import numpy as np
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_random_exponential,
)

import outlines
from outlines.caching import cache

__all__ = ["OpenAIAPI", "openai"]


class OpenAIAPI:
    def __init__(
        self,
        model_name: str,
        api_key: Optional[str] = os.getenv("OPENAI_API_KEY"),
        temperature: float = 1.0,
    ):
        self.api_key = api_key

        if "text-" in model_name:
            call_api = call_completion_api
            format_prompt = lambda x: x
            extract_choice = lambda x: x["text"]
        elif "gpt-" in model_name:
            call_api = call_chat_completion_api
            format_prompt = lambda x: [{"role": "user", "content": x}]
            extract_choice = lambda x: x["message"]["content"]
        else:
            raise NameError(
                f"The model {model_name} requested is not available. Only the completion and chat completion models are available for OpenAI."
            )

        @functools.partial(outlines.vectorize, signature="(),(),(m),(),()->(s)")
        async def generate_base(
            prompt: str,
            max_tokens: int,
            stop_at: List[Optional[str]],
            samples: int,
            api_key: str,
        ) -> str:
            responses = await call_api(
                model_name,
                format_prompt(prompt),
                int(max_tokens),
                temperature,
                stop_at,
                {},
                samples,
                api_key,
            )

            if samples == 1:
                results = np.array([extract_choice(responses["choices"][0])])
            else:
                results = np.array(
                    [extract_choice(responses["choices"][i]) for i in range(samples)]
                )

            return results

        @functools.partial(outlines.vectorize, signature="(),(),(m),(),()->(s)")
        async def generate_choice(
            prompt: str, max_tokens: int, is_in: List[str], samples: int, api_key: str
        ) -> Union[List[str], str]:
            """Generate a sequence that must be one of many options.

            .. warning::

                This function will call the API once for every token generated.

            We tokenize every choice, iterate over the token lists, create a mask
            with the current tokens and generate one token. We progressively
            eliminate the choices that don't start with the currently decoded
            sequence.

            """
            try:
                import tiktoken
            except ImportError:
                raise ImportError(
                    "The `tiktoken` library needs to be installed in order to choose `outlines.models.openai` with `is_in`"
                )

            tokenizer = tiktoken.encoding_for_model(model_name)
            encoded: List[List[int]] = [tokenizer.encode(word) for word in is_in]

            decoded_samples = []
            for _ in range(samples):
                decoded: List[str] = []
                for i in range(max([len(word) for word in encoded])):
                    mask = {}
                    for word, tokenized_word in zip(is_in, encoded):
                        if not word.startswith("".join(decoded)):
                            continue
                        try:
                            mask[tokenized_word[i]] = 100
                        except IndexError:
                            pass

                    if len(mask) == 0:
                        break

                    response = await call_api(
                        model_name,
                        format_prompt(prompt),
                        1,
                        temperature,
                        [],
                        mask,
                        samples,
                        api_key,
                    )
                    decoded.append(extract_choice(response["choices"][0]))
                    prompt = prompt + "".join(decoded)

                decoded_samples.append("".join(decoded))

            return np.array(decoded_samples)

        self.generate_base = generate_base
        self.generate_choice = generate_choice

    def __call__(
        self,
        prompt: str,
        max_tokens: int = 500,
        *,
        samples=1,
        stop_at: Union[List[Optional[str]], str] = [],
        is_in: Optional[List[str]] = None,
    ):
        if is_in is not None and stop_at:
            raise TypeError("You cannot set `is_in` and `stop_at` at the same time.")
        elif is_in is not None:
            return self.generate_choice(
                prompt, max_tokens, is_in, samples, self.api_key
            )
        else:
            if isinstance(stop_at, str):
                stop_at = [stop_at]
            return self.generate_base(
                prompt, max_tokens, stop_at, samples, self.api_key
            )


openai = OpenAIAPI


def error_handler(api_call_fn: Callable) -> Callable:
    """Handle OpenAI API errors and missing API key."""

    def call(*args, **kwargs):
        import openai

        try:
            return api_call_fn(*args, **kwargs)
        except (
            openai.error.RateLimitError,
            openai.error.Timeout,
            openai.error.TryAgain,
            openai.error.APIConnectionError,
            openai.error.ServiceUnavailableError,
        ) as e:
            raise OSError(f"Could not connect to the OpenAI API: {e}")
        except (
            openai.error.AuthenticationError,
            openai.error.PermissionError,
            openai.error.InvalidRequestError,
            openai.error.InvalidAPIType,
        ) as e:
            raise e

    return call


retry_config = {
    "wait": wait_random_exponential(min=1, max=30),
    "stop": stop_after_attempt(6),
    "retry": retry_if_exception_type(OSError),
}


@retry(**retry_config)
@error_handler
@cache
async def call_completion_api(
    model: str,
    prompt: str,
    max_tokens: int,
    temperature: float,
    stop_sequences: List[str],
    logit_bias: Dict[str, int],
    num_samples: int,
    api_key: str,
):
    try:
        import openai
    except ImportError:
        raise ImportError(
            "The `openai` library needs to be installed in order to use Outlines' OpenAI integration."
        )

    response = await openai.Completion.acreate(
        engine=model,
        prompt=prompt,
        temperature=temperature,
        max_tokens=max_tokens,
        stop=list(stop_sequences) if len(stop_sequences) > 0 else None,
        logit_bias=logit_bias,
        n=int(num_samples),
        api_key=api_key,
    )
    return response


@retry(**retry_config)
@error_handler
@cache
async def call_chat_completion_api(
    model: str,
    messages: List[Dict[str, str]],
    max_tokens: int,
    temperature: float,
    stop_sequences: List[str],
    logit_bias: Dict[str, int],
    num_samples: int,
    api_key: str,
):
    try:
        import openai
    except ImportError:
        raise ImportError(
            "The `openai` library needs to be installed in order to use Outlines' OpenAI integration."
        )

    response = await openai.ChatCompletion.acreate(
        model=model,
        messages=messages,
        max_tokens=max_tokens,
        temperature=temperature,
        stop=list(stop_sequences) if len(stop_sequences) > 0 else None,
        logit_bias=logit_bias,
        n=int(num_samples),
        api_key=api_key,
    )

    return response



import json as pyjson
import math
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Set, Tuple, Union

import interegular
import torch
from pydantic import BaseModel

from outlines.text.fsm import create_fsm_index_tokenizer, make_deterministic_fsm
from outlines.text.generate.continuation import Continuation
from outlines.text.json_schema import build_regex_from_object, get_schema_from_signature
from outlines.text.types import python_types_to_regex

if TYPE_CHECKING:
    from outlines.text.generate.sample import Sampler


class Regex(Continuation):
    """Represents a regex-based generation model.

    `Regex` instances are constrained generation models that only generate
    sequences matching a given regex.

    >>> import outlines.text as text
    >>> generator = text.generate.regex(model, "(0|[1-9][0-9]+)")

    Sequences can then be generated from a prompt as follows:

    >>> sequence_1 = generator("Return an integer between 0 and 10")
    >>> sequence_2 = generator("Rate the movie "Hackers" on a scale from 0 to 10")

    .. note:
        Reuse instances of these guided generators (e.g. `generator` from the
        above example) whenever possible, because constructing them has more
        overhead than generating token sequences from them.

    """

    def __init__(
        self,
        model,
        regex_string: str,
        max_tokens: Optional[int] = None,
        *,
        sampler: Optional["Sampler"] = None,
        stop: Union[str, List[str]] = [],
        allow_empty_tokens: bool = True,
        initial_state: Optional[int] = None,
        final_states: Optional[Set[int]] = None,
        states_to_token_maps: Optional[Dict[int, Dict[int, int]]] = None,
        empty_token_ids: Optional[Set[int]] = None,
        format_fn: Callable[[str], Union[BaseModel, dict, str]] = lambda x: x,
    ):
        """

        Parameters
        ----------
        model
            The instance of the model used to generate next-token probabilities.
        regex_string
            The regex with which the token sampling process is guided/constrained.
        max_tokens
            The maximum number of tokens to be sampled.
        sampler
            The function used to draw samples.  Defaults to
            `outlines.text.generate.sample.multinomial`.  See
            `outlines.text.generate.sample.Sampler` for the expected form of
            such functions.
        stop
            Optional stopping string(s).
        allow_empty_tokens
            Allow sampling of tokens corresponding to empty strings.
        states_to_token_maps
            Pre-computed map of FSM start states to maps between token ids and their
            corresponding FSM end states.
        empty_token_ids
            Pre-computed set of token ids for tokens that are empty strings.
        format_fn
            The function to apply to the generated JSON.

        """
        super().__init__(model, max_tokens, sampler, stop)

        if (
            states_to_token_maps is None
            or empty_token_ids is None
            or initial_state is None
            or final_states is None
        ):
            regex_pattern = interegular.parse_pattern(regex_string)
            regex_fsm, _ = make_deterministic_fsm(regex_pattern.to_fsm().reduce())

            (
                self.states_to_token_maps,
                self.empty_token_ids,
            ) = create_fsm_index_tokenizer(regex_fsm, model.tokenizer)
            self.initial_state = regex_fsm.initial
            self.final_states = regex_fsm.finals
        else:
            self.initial_state = initial_state
            self.final_states = final_states
            self.states_to_token_maps = states_to_token_maps
            self.empty_token_ids = empty_token_ids

        # Check whether a terminal path (from the initial state of the FSM to
        # one of its terminal states) exists, raise an exception otherwise.
        if not any(
            self.final_states.intersection(v.values())
            for v in self.states_to_token_maps.values()
        ):
            raise ValueError(
                "The vocabulary does not allow us to build a sequence that matches the input regex"
            )

        # When an EOS is observed, the last FSM state becomes `-1`.
        self.last_fsm_states: List[int] = []
        self.mask_cache: Dict[Tuple[int, int], torch.LongTensor] = {}
        self.regex_string = regex_string
        self.allow_empty_tokens = allow_empty_tokens
        self.format_fn = format_fn

    def create_proposal(
        self, generated_token_ids: torch.LongTensor, logits: torch.DoubleTensor
    ) -> torch.DoubleTensor:
        """Modify the next-token logits so that only valid tokens can be generated.

        Parameters
        ----------
        generated_token_ids
            The token ids generated so far.
        logits
            The next-token logits.

        """

        assert generated_token_ids.ndim == 2

        if len(self.last_fsm_states) == 0:
            self.last_fsm_states = [self.initial_state for _ in range(logits.shape[0])]

        masks = []

        for i, (token_seq, last_state) in enumerate(
            zip(
                generated_token_ids,
                self.last_fsm_states,
            )
        ):
            if token_seq.shape[0] > 0:
                # Get the last token that was sampled
                last_token = int(token_seq[-1])

                if last_token in self.empty_token_ids:
                    # An empty token was sampled, so the FSM state hasn't changed
                    next_state = last_state
                    next_token_ids = list(self.states_to_token_maps[last_state].keys())

                elif last_token != self.model.tokenizer.eos_token_id:
                    # If we previously ended with an EOS, we shouldn't be
                    # getting/sampling any more non-EOS tokens.
                    assert last_state > -1

                    last_token_to_end_state = self.states_to_token_maps[last_state]

                    next_state = last_token_to_end_state[last_token]

                    next_tokens_to_end_states = self.states_to_token_maps.get(
                        next_state
                    )

                    if next_tokens_to_end_states is None:
                        # If there are no transitions from the current state,
                        # then we must've been in a final state of the FSM.
                        # We produce EOS tokens from here on.
                        assert next_state in self.final_states
                        next_state = -1
                        next_token_ids = [self.model.tokenizer.eos_token_id]
                    else:
                        next_token_ids = list(next_tokens_to_end_states.keys())
                else:
                    # Since we already have an EOS, only sample EOS tokes from
                    # here on.
                    next_state = -1
                    next_token_ids = [self.model.tokenizer.eos_token_id]
            else:
                # There weren't any previous tokens, so we can't update the state
                next_state = last_state
                next_token_ids = list(self.states_to_token_maps[last_state].keys())

            mask = self._get_mask_for_state(
                next_state, logits.shape[-1], next_token_ids
            )
            masks.append(mask)
            self.last_fsm_states[i] = next_state

        mask = torch.concatenate(masks, dim=0)

        return logits + mask

    def _get_mask_for_state(
        self, state: int, size: int, next_token_ids: List[int]
    ) -> torch.LongTensor:
        mask = self.mask_cache.get((state, size))

        if mask is None:
            mask = torch.full(
                (size,),
                -math.inf,
                device=self.device,
            )

            if self.allow_empty_tokens:
                token_ids = list(self.empty_token_ids) + next_token_ids
            else:
                token_ids = next_token_ids

            mask[token_ids] = 0
            mask = mask.unsqueeze(0)
            self.mask_cache[(state, size)] = mask

        return mask

    def postprocess_completions(self, completions: List[str]):
        self.last_fsm_states.clear()
        results: List[str] = super().postprocess_completions(completions)
        return [self.format_fn(result) for result in results]


def regex(
    model,
    regex_string: str,
    max_tokens: Optional[int] = None,
    *,
    sampler: Optional["Sampler"] = None,
    allow_empty_tokens: bool = True,
):
    """Generate text sequences that match the input regex.

    .. note:
        Reuse instances of these guided generators whenever possible,
        because constructing them has more overhead than generating
        token sequences from them.  See the docstring for `Regex`.

    Parameters
    ----------
    model
        The language model to use to compute the next-token logits.
    regex_string
        The regular expression that generated expressions must match.
    max_tokens
        The maximum number of tokens to generate.
    sampler
        The function used to draw samples.  Defaults to
        `outlines.text.generate.sample.multinomial`.  See
        `outlines.text.generate.sample.Sampler` for the expected form of
        such functions.
    allow_empty_tokens
        Allow sampling of tokens corresponding to empty strings.

    """
    return Regex(
        model,
        regex_string,
        max_tokens,
        sampler=sampler,
        allow_empty_tokens=allow_empty_tokens,
    )


def format(
    model,
    python_type,
    max_tokens: Optional[int] = None,
    *,
    sampler: Optional["Sampler"] = None,
    allow_empty_tokens: bool = True,
):
    """Generate integers.

    The regex used to constrain the generation optionally matches plus or minus
    signs and forbids leading zeros (even if the `int` function in Python allows
    them).

    .. note:
        Reuse instances of these guided generators whenever possible,
        because constructing them has more overhead than generating
        token sequences from them.  See the docstring for `Regex`.

    Parameters
    ----------
    model
        The language model to use to compute the next-token logits.
    python_type
        The format in which the output is expected, defined as a Python type.
    max_tokens
        The maximum number of tokens to generate.
    sampler
        The function used to draw samples.  Defaults to
        `outlines.text.generate.sample.multinomial`.  See
        `outlines.text.generate.sample.Sampler` for the expected form of
        such functions.
    allow_empty_tokens
        Allow sampling of tokens corresponding to empty strings.

    """
    regex_str = python_types_to_regex(python_type)
    return Regex(
        model,
        regex_str,
        max_tokens,
        sampler=sampler,
        allow_empty_tokens=allow_empty_tokens,
    )


def choice(
    model,
    choices: List[str],
    max_tokens: Optional[int] = None,
    *,
    sampler: Optional["Sampler"] = None,
    allow_empty_tokens: bool = True,
):
    """Choose between different sequences.

    .. note:
        Reuse instances of these guided generators whenever possible,
        because constructing them has more overhead than generating
        token sequences from them.  See the docstring for `Regex`.

    Parameters
    ----------
    model
        The language model to use to compute the next-token logits.
    max_tokens
        The maximum number of tokens to generate.
    sampler
        The function used to draw samples.  Defaults to
        `outlines.text.generate.sample.multinomial`.  See
        `outlines.text.generate.sample.Sampler` for the expected form of
        such functions.
    allow_empty_tokens
        Allow sampling of tokens corresponding to empty strings.
    """
    regex_str = r"(" + r"|".join(choices) + r")"
    return Regex(
        model,
        regex_str,
        max_tokens,
        sampler=sampler,
        allow_empty_tokens=allow_empty_tokens,
    )


def json(
    model,
    schema_object: Union[str, BaseModel, Callable],
    max_tokens: Optional[int] = None,
    *,
    sampler: Optional["Sampler"] = None,
    allow_empty_tokens: bool = True,
) -> Union[dict, BaseModel]:
    """Generate a text sequence that follows a JSON schema or Pydantic model.

    .. note:
        Reuse instances of these guided generators whenever possible,
        because constructing them has more overhead than generating
        token sequences from them. See the docstring for `Regex`.

    Parameters
    ---------
    model
        The language model to use to compute the next-token logits.
    schema
        The JSON schema, Pydantic model or function (signature) that guides the
        generation.
    max_tokens
        The maximum number of tokens to generate.
    sampler
        The function used to draw samples.  Defaults to
        `outlines.text.generate.sample.multinomial`.  See
        `outlines.text.generate.sample.Sampler` for the expected form of
        such functions.
    allow_empty_tokens
        Allow sampling of tokens corresponding to empty strings.

    """
    if isinstance(schema_object, type(BaseModel)):
        schema = pyjson.dumps(schema_object.model_json_schema())
        format_fn = lambda x: schema_object.model_validate(pyjson.loads(x))
    elif callable(schema_object):
        schema = pyjson.dumps(get_schema_from_signature(schema_object))
        # TODO: Convert string fields to their respective types
        format_fn = lambda x: pyjson.loads(x)
    else:
        format_fn = lambda x: x

    regex_str = build_regex_from_object(schema)

    return Regex(
        model,
        regex_str,
        max_tokens,
        sampler=sampler,
        allow_empty_tokens=allow_empty_tokens,
        format_fn=format_fn,
    )



import torch

from outlines.text.generate.continuation import Continuation, continuation


class Tokenizer:
    eos_token = "<EOS>"
    eos_token_id = 0
    pad_token_id = -1

    def decode(self, token_ids):
        return ["Test"] * token_ids.shape[0]


class Model:
    tokenizer = Tokenizer()
    device = "cpu"


def test_continuation_eos_is_finished():
    model = continuation(Model())
    assert isinstance(model, Continuation)

    token_ids = torch.tensor([[3, 2]])
    result = model.is_finished(token_ids)
    assert torch.equal(result, torch.tensor([False]))

    token_ids = torch.tensor([[3, 2, 0]])
    result = model.is_finished(token_ids)
    assert torch.equal(result, torch.tensor([True]))

    token_ids = torch.tensor([[3, 2, 1], [3, 2, 0]])
    result = model.is_finished(token_ids)
    assert torch.equal(result, torch.tensor([False, True]))

    token_ids = torch.tensor([[3, 2, 1, 0], [3, 2, 0, -1]])
    result = model.is_finished(token_ids)
    assert torch.equal(result, torch.tensor([True, False]))


def test_continuation_postprocess():
    model = continuation(Model())
    result = model.postprocess_completions(["Here<EOS>"])
    assert len(result) == 1
    assert result[0] == "Here"


def test_continuation_stop_is_finished():
    tokenizer = Tokenizer()
    tokenizer.decode = lambda x: ["finished \n", "not_finished"]
    model = Model()
    model.tokenizer = tokenizer

    model = continuation(model, stop=["\n"])

    token_ids = torch.tensor([[2, 3], [2, 3]])
    result = model.is_finished(token_ids)
    assert torch.equal(result, torch.tensor([True, False]))


def test_continuation_stop_postprocess():
    model = Continuation(Model(), stop="\n")
    result = model.postprocess_completions(["Stop\n"])
    assert len(result) == 1
    assert result[0] == "Stop"

    model = Continuation(Model(), stop=["\n", ","])
    result = model.postprocess_completions(["Stop"])
    assert len(result) == 1
    assert result[0] == "Stop"

    result = model.postprocess_completions(["Stop\n"])
    assert len(result) == 1
    assert result[0] == "Stop"

    result = model.postprocess_completions(["Stop\naaa"])
    assert len(result) == 1
    assert result[0] == "Stop"

    result = model.postprocess_completions(["Stop,aa\naaa"])
    assert len(result) == 1
    assert result[0] == "Stop"

    result = model.postprocess_completions(["Stop\naa,a"])
    assert len(result) == 1
    assert result[0] == "Stop"

    result = model.postprocess_completions(["Stop\n", "Nonstop"])
    assert len(result) == 2
    assert result == ["Stop", "Nonstop"]

    result = model.postprocess_completions(["StopHere\nNoHere<EOS>"])
    assert len(result) == 1
    assert result[0] == "StopHere"



import math
from typing import Dict, List, Union

import numpy as np
import pytest
import torch

from outlines import models
from outlines.models.tokenizer import Tokenizer
from outlines.text.generate.sequence import Sequence


def test_openai_error():
    model = models.openai("text-davinci-003")
    with pytest.raises(TypeError):
        Sequence(model)


class MockModel:
    def __init__(self, tokenizer, logits):
        self.tokenizer = tokenizer
        self.logits = logits
        self.iteration_idx = 0
        self.device = "cpu"

    def forward(self, input_ids, *_):
        import math

        batch_shape = input_ids.shape[:-1]
        vocab_shape = (self.logits.shape[-1],)
        shaped_logits = torch.tile(
            self.logits[self.iteration_idx], (math.prod(batch_shape), 1)
        )
        self.iteration_idx += 1

        return shaped_logits.reshape(batch_shape + vocab_shape), None

    def __call__(self, input_ids, *_):
        return self.forward(input_ids)[0]


class MockTokenizer(Tokenizer):
    def __init__(self, vocabulary: Dict[str, int]):
        self.vocabulary = vocabulary
        self.id_to_str = {v: k for k, v in vocabulary.items()} if vocabulary else {}
        self.pad_token_id = -1
        self.id_to_str[self.pad_token_id] = "<pad>"

    def encode(self, prompts: Union[str, List[str]]):
        if isinstance(prompts, str):
            prompts = [prompts]

        token_ids = torch.tensor([[self.vocabulary[prompt]] for prompt in prompts])
        attention_mask = torch.ones_like(token_ids)

        return token_ids, attention_mask

    def decode(self, token_ids):
        ndims = np.ndim(token_ids)

        assert 0 < ndims <= 2

        if ndims == 1:
            token_ids = [token_ids]

        res = ["".join(self.id_to_str[int(idx)] for idx in seq) for seq in token_ids]

        return res if ndims > 1 else res[0]

    def convert_token_to_string(self, token: str) -> str:
        return token

    def __hash__(self):
        return id(self)


def test_sequence_error():
    with pytest.raises(NotImplementedError, match="must be implemented"):
        sequence = Sequence(MockModel(MockTokenizer(None), None))
        sequence.is_finished(torch.tensor([1]))


class ModelStep:
    """Mock model to test `Sequence.step`"""

    def __init__(self, tokenizer, logits):
        self.device = "cpu"
        self.logits = logits
        self.tokenizer = tokenizer

    def forward(self, input_ids, *_):
        """Call the model.

        We first repeat the logits `num_sequences` times, and then
        reshape the resulting array to match the batch size.

        """
        import math

        batch_shape = input_ids.shape[:-1]
        vocab_shape = (self.logits.shape[-1],)
        shaped_logits = torch.tile(self.logits, (math.prod(batch_shape), 1))
        return shaped_logits.reshape(batch_shape + vocab_shape), None

    def __call__(self, input_ids, *_):
        return self.forward(input_ids)[0]


def test_sequence_step():
    rng = torch.Generator()
    rng.manual_seed(0)

    logits = torch.tensor([-math.inf, 1, -math.inf, -math.inf], dtype=torch.double)
    model = ModelStep(MockTokenizer(None), logits)

    sequence = Sequence(model)

    input_ids = torch.tensor([[1, 2]])
    token_ids, probs, _ = sequence.step(rng, 2, input_ids, torch.ones((1, 2)))
    assert torch.equal(token_ids, torch.tensor([[[1]]]))
    assert probs.shape == (1, 1, 4)


def test_sequence_step_batch():
    rng = torch.Generator()
    rng.manual_seed(0)

    logits = torch.tensor([-math.inf, 0.5, 0.5, -math.inf], dtype=torch.double)
    model = ModelStep(MockTokenizer(None), logits)

    sequence = Sequence(model)

    input_ids = torch.tensor([[1, 2], [3, 4]])
    token_ids, probs, _ = sequence.step(rng, 2, input_ids, torch.ones((2, 2)))
    assert torch.equal(token_ids, torch.tensor([[[1], [2]]]))
    assert probs.shape == (1, 2, 4)


def test_sequence_step_sample():
    rng = torch.Generator()
    rng.manual_seed(0)

    logits = torch.tensor([-math.inf, 0.5, 0.5, -math.inf], dtype=torch.double)
    model = ModelStep(MockTokenizer(None), logits)

    sequence = Sequence(model)
    input_ids = torch.tensor([[1, 2]])
    token_ids, probs, _ = sequence.step(
        rng, 2, input_ids, torch.ones((1, 2)), samples=3
    )
    assert torch.equal(token_ids, torch.tensor([[[1]], [[2]], [[1]]]))
    assert probs.shape == (3, 1, 4)


def test_sequence_step_sample_batch():
    rng = torch.Generator()
    rng.manual_seed(0)

    logits = torch.tensor([-math.inf, 0.5, 0.5, -math.inf], dtype=torch.double)
    model = ModelStep(MockTokenizer(None), logits)

    sequence = Sequence(model)
    input_ids = torch.tensor([[1, 2, 1], [3, 4, 1]])
    token_ids, probs, _ = sequence.step(
        rng, 3, input_ids, torch.ones((2, 3)), samples=3
    )
    assert torch.equal(
        token_ids,
        torch.tensor(
            [
                [[1], [2]],
                [[1], [1]],
                [[1], [2]],
            ]
        ),
    )
    assert probs.shape == (3, 2, 4)


def test_sequence_step_loop():
    """Make sure that we can feed `step`'s output back as an input."""
    rng = torch.Generator()
    rng.manual_seed(0)

    logits = torch.tensor([-math.inf, 0.5, 0.5, -math.inf], dtype=torch.double)
    model = ModelStep(MockTokenizer(None), logits)

    sequence = Sequence(model)
    input_ids = torch.tensor([[1, 2]])
    token_ids, *_ = sequence.step(rng, 2, input_ids, torch.ones((1, 2)))
    token_ids, probs, _ = sequence.step(
        rng, 2, token_ids.squeeze(0), torch.ones((1, 3))
    )
    assert torch.equal(token_ids, torch.tensor([[[2]]]))
    assert probs.shape == (1, 1, 4)

    input_ids = torch.tensor([[1, 2], [3, 4]])
    token_ids, *_ = sequence.step(rng, 2, input_ids, torch.ones((2, 2)))
    token_ids, probs, _ = sequence.step(
        rng, 2, token_ids.squeeze(0), torch.ones((2, 3))
    )
    assert torch.equal(token_ids, torch.tensor([[[1], [2]]]))
    assert probs.shape == (1, 2, 4)

    # The number of samples becomes the batch size at the next iteration.
    input_ids = torch.tensor([[1, 2]])
    token_ids, *_ = sequence.step(rng, 2, input_ids, torch.ones((1, 2)), samples=3)
    token_ids, probs, _ = sequence.step(
        rng, 2, token_ids.squeeze(1), torch.ones((3, 3))
    )
    assert torch.equal(token_ids, torch.tensor([[[2], [1], [1]]]))
    assert probs.shape == (1, 3, 4)


def test_sequence_step_loop_general():
    rng = torch.Generator()
    rng.manual_seed(0)

    logits = torch.tensor([-math.inf, 0.5, 0.5, -math.inf], dtype=torch.double)
    model = ModelStep(MockTokenizer(None), logits)

    sequence = Sequence(model)
    input_ids = torch.tensor([[1, 2, 1], [3, 4, 1]])
    token_ids, *_ = sequence.step(rng, 3, input_ids, torch.ones((1, 3)), samples=3)
    result, *_ = sequence.step(rng, 3, token_ids, torch.ones((3, 4)))
    assert result.shape == (1, 3, 2, 1)
    assert torch.equal(
        result.squeeze(0),
        torch.tensor(
            [
                [[1], [2]],
                [[1], [2]],
                [[1], [1]],
            ]
        ),
    )


def test_call_single_prompt():
    class FinishAfterTwo(Sequence):
        def __init__(self, model):
            super().__init__(model)
            self.iteration_idx = 0

        def is_finished(self, token_ids):
            """Finish generating the sequence after two iterations"""
            if self.iteration_idx == 0:
                self.iteration_idx += 1
                return torch.tensor([False])
            else:
                return torch.tensor([True])

    tokenizer = MockTokenizer({"Test": 0, "a": 1, "b": 2})
    model = MockModel(
        tokenizer,
        torch.tensor([[1.0, -math.inf, -math.inf], [-math.inf, 1.0, -math.inf]]),
    )
    sequence = FinishAfterTwo(model)

    result = sequence("Test")
    assert result == "Testa"


def test_call_prompt_list():
    class FinishAfterThree(Sequence):
        def __init__(self, model):
            super().__init__(model)
            self.iteration_idx = 0

        def is_finished(self, token_ids):
            """Finish generating the first sequence after two iteration and the
            second one after two iterations.

            """
            if self.iteration_idx == 0:
                self.iteration_idx += 1
                return torch.tensor([False, False, False])
            elif self.iteration_idx == 1:
                self.iteration_idx += 1
                return torch.tensor([True, False, True])
            else:
                return torch.tensor([True])  # We only consider the unfinished sequences

    tokenizer = MockTokenizer(
        {"Test1": 0, "Test2": 1, "a": 2, "b": 3, "c": 4, "Test3": 5}
    )
    model = MockModel(
        tokenizer,
        torch.tensor(
            [
                [-math.inf, -math.inf, 1.0, -math.inf, -math.inf, -math.inf],
                [-math.inf, -math.inf, -math.inf, 1.0, -math.inf, -math.inf],
                [-math.inf, -math.inf, -math.inf, -math.inf, 1.0, -math.inf],
            ]
        ),
    )
    sequence = FinishAfterThree(model)

    result = sequence(["Test1", "Test2", "Test3"])
    assert result == ["ab<pad>", "abc", "ab<pad>"]


def test_call_single_prompt_samples():
    class FinishAfterTwo(Sequence):
        def __init__(self, model):
            super().__init__(model)
            self.iteration_idx = 0

        def is_finished(self, token_ids):
            if self.iteration_idx == 0:
                self.iteration_idx += 1
                return torch.tensor([False, False, False])
            else:
                return torch.tensor([True, True, True])

    tokenizer = MockTokenizer({"a": 0, "b": 1, "c": 2, "Test": 4})
    model = MockModel(
        tokenizer,
        torch.tensor(
            [
                [1, -math.inf, -math.inf, -math.inf],
                [-math.inf, 1, -math.inf, -math.inf],
            ],
            dtype=torch.double,
        ),
    )
    sequence = FinishAfterTwo(model)
    result = sequence("Test", samples=3)
    assert result == ["ab", "ab", "ab"]

    class FinishAfterOne(Sequence):
        def __init__(self, model):
            super().__init__(model)

        def is_finished(self, token_ids):
            return torch.tensor([True, True, True])

    tokenizer = MockTokenizer({"a": 0, "b": 1, "c": 3, "Test": 4})
    model = MockModel(
        tokenizer,
        torch.tensor(
            [
                [1, -math.inf, -math.inf, -math.inf],
                [-math.inf, 1, -math.inf, -math.inf],
            ],
            dtype=torch.double,
        ),
    )
    sequence = FinishAfterOne(model)
    result = sequence("Test", samples=3)
    assert result == ["a", "a", "a"]


def test_call_prompt_list_samples():
    class FinishAfterThree(Sequence):
        def __init__(self, model):
            super().__init__(model)
            self.iteration_idx = 0

        def is_finished(self, token_ids):
            if self.iteration_idx == 0:
                self.iteration_idx += 1
                batch_shape = token_ids.shape[:-1]
                return torch.zeros(batch_shape, dtype=torch.bool)
            elif self.iteration_idx == 1:
                self.iteration_idx += 1
                return torch.tensor(
                    [[True, False, True], [True, False, True], [True, False, True]]
                )
            else:
                return torch.tensor([True, True, True])

    tokenizer = MockTokenizer(
        {"a": 0, "b": 1, "c": 2, "Test1": 3, "Test2": 4, "Test3": 5}
    )
    model = MockModel(
        tokenizer,
        torch.tensor(
            [
                [1, -math.inf, -math.inf, -math.inf, -math.inf, -math.inf],
                [-math.inf, 1, -math.inf, -math.inf, -math.inf, -math.inf],
                [-math.inf, -math.inf, 1, -math.inf, -math.inf, -math.inf],
            ],
            dtype=torch.double,
        ),
    )
    sequence = FinishAfterThree(model)

    result = sequence(["Test1", "Test2", "Test3"], samples=3)
    assert result == [
        "ab<pad>",
        "abc",
        "ab<pad>",
        "ab<pad>",
        "abc",
        "ab<pad>",
        "ab<pad>",
        "abc",
        "ab<pad>",
    ]



from typing import Protocol

import torch


class Sampler(Protocol):
    def __call__(
        self, logits: torch.DoubleTensor, samples: int, rng: torch.Generator
    ) -> torch.DoubleTensor:
        ...


def greedy(logits: torch.DoubleTensor, samples: int, *_) -> torch.DoubleTensor:
    """Greedy Sampling algorithm.

    Greedy sampling consists in choosing the token with the largest
    likelihood at every step.

    Parameters
    ----------
    logits
        A tensor of shape ``(n_seqs, vocab_size,)`` that represents the
        probability distribution of the next token over the vocabulary.
    samples
        The number of sequences to produce.  In this case, the top-`samples`
        logit values are returned.
    rng
        A random number generator.

    Returns
    -------
    The ids of the sampled tokens having shape ``(samples, n_seqs)``.

    """
    if samples == 1:
        next_token_ids = torch.argmax(logits, dim=-1, keepdim=True).T
    else:
        next_token_ids = torch.topk(
            logits, samples, dim=-1, largest=True, sorted=True
        ).indices.T

    return next_token_ids


def multinomial(
    logits: torch.DoubleTensor, samples: int, rng: torch.Generator
) -> torch.DoubleTensor:
    """Multinomial sampling algorithm.

    Multinomial sampling consists in randomly sampling the next token assuming
    its distribution is a Categorical distribution parametrized by the
    next-token logits.

    Parameters
    ----------
    logits
        A tensor of shape ``(n_seqs, vocab_size,)`` that represents the
        probability distribution of the next token over the vocabulary.
    samples
        The number of sequences to sample.
    rng
        A random number generator.

    Returns
    -------
    The ids of the sampled tokens having shape ``(samples, n_seqs)``.

    """
    probs = torch.nn.functional.softmax(logits, dim=-1)
    # next_token_ids = torch.multinomial(probs, num_samples=samples, generator=rng)
    next_token_ids = vectorized_random_choice(rng, probs, samples)
    return next_token_ids


def vectorized_random_choice(
    rng: torch.Generator,
    p: torch.FloatTensor,
    samples: int = 1,
):
    """Vectorized implementation of `np.random.choice`.

    `np.random.choice` does not support arrays of probability. This implements
    the equivalent of this function where the `p` argument can be a matrix.

    Note
    ----
    `torch.searchsorted` may be more efficient, but it is not implemented for
    every backend, for instance MPS.

    Parameters
    ----------
    rng
        Torch random number Generator instance
    p
        An array of probability of shape ``(num_probability_vectors, num_items)``
        that must sum to 1.
    samples
        The number of samples to take for each probability vector.

    Returns
    -------
    An array of shape ``(num_samples, batch_size)``

    """
    cumsum = torch.unsqueeze(p.cumsum(axis=-1), 0)
    rand = torch.rand(
        (samples,) + p.shape[:-1] + (1,), generator=rng, device=rng.device
    )
    idx = (cumsum < rand).sum(axis=-1)

    return idx



from typing import TYPE_CHECKING, List, Optional, Union

import torch

from outlines.text.generate.sequence import Sequence

if TYPE_CHECKING:
    from outlines.text.generate.sample import Sampler


class Continuation(Sequence):
    """Represents a completion generation model.

    `Continuation` instances are unconstrained generation models that stop when
    an EOS token has been found or when the maximum number of tokens has been
    reached.

    >>> import outlines.text as text
    >>> sequence = text.generate.continuation(model)("Say something")

    """

    def __init__(
        self,
        model,
        max_tokens: Optional[int] = None,
        sampler: Optional["Sampler"] = None,
        stop: Union[str, List[str]] = [],
    ):
        super().__init__(model, max_tokens, sampler)
        self.eos_token_id = torch.tensor(
            [self.model.tokenizer.eos_token_id], device=self.device
        )

        if isinstance(stop, str):
            stop = [stop]

        self.stop_sequences = stop

    def is_finished(self, token_ids: torch.LongTensor) -> torch.BoolTensor:
        """Determine whether the sequences reached maximum length of end with
        and EOS token.

        We only need to look for the EOS token in the last element rather than
        in the whole sequence. Indeed, (1) EOS is a single token (2)
        `Sequence`'s `__call__` methods only passed the `token_ids` of the
        sequences that haven't been marked as finished already.

        Parameters
        ----------
        token_ids
            The input sequences.

        """

        contains_eos = token_ids[:, -1] == self.model.tokenizer.eos_token_id

        if self.stop_sequences:
            sequences = self.model.tokenizer.decode(token_ids)
            contains_stop_sequence = []
            for sequence in sequences:
                contains_stop_sequence.append(
                    any(stop_str in sequence for stop_str in self.stop_sequences)
                )
            contains_stop_sequence = torch.tensor(
                contains_stop_sequence, dtype=torch.bool, device=self.model.device
            )

            return torch.logical_or(contains_eos, contains_stop_sequence)
        else:
            return contains_eos

    def postprocess_completions(self, completions: List[str]) -> List[str]:
        """Remove the EOS token from the completion.

        Sequences in `stop` take precedence over EOS. For instance, if
        `stop=["\n"]` and the generated sequence is 'One\nTwo<EOS>`
        `Continuation.postprocess_completions` will return `One`.

        """
        completions_without_eos = [
            completion.replace(self.model.tokenizer.eos_token, "")
            for completion in completions
        ]

        completions_without_stop = []
        for completion in completions_without_eos:
            for stop_str in self.stop_sequences:
                idx = completion.rfind(stop_str)  # ignore the prompt
                if idx > 0:
                    completion = completion[:idx]

            completions_without_stop.append(completion)

        return completions_without_stop


def continuation(
    model,
    max_tokens: Optional[int] = None,
    *,
    sampler: Optional["Sampler"] = None,
    stop: Union[str, List[str]] = [],
):
    """Generate text sequences.

    Parameters
    ----------
    model
        The language model to use to compute the next-token logits.
    max_tokens
        The maximum number of tokens to generate.
    sampler
        The function used to draw samples.  Defaults to
        `outlines.text.generate.sample.multinomial`.  See
        `outlines.text.generate.sample.Sampler` for the expected form of
        such functions.
    stop
        A string or list of strings which, when generated, stops
        the generation for this sequence.

    """
    return Continuation(model, max_tokens, sampler, stop)



import math

import torch

from outlines.text.generate.sample import greedy, multinomial, vectorized_random_choice


def test_greedy():
    logits = torch.tensor([[1.0, 2.0, 5.0]])
    next_token_ids = greedy(logits, samples=1)
    assert next_token_ids.equal(torch.tensor([[2]]))

    next_token_ids = greedy(logits, samples=2)
    assert next_token_ids.equal(torch.tensor([[2], [1]]))

    logits = torch.tensor([[10.0, 0.0, 3.0], [-math.inf, 2.0, 5.0]])
    next_token_ids = greedy(logits, samples=1)
    assert next_token_ids.equal(torch.tensor([[0, 2]]))

    next_token_ids = greedy(logits, samples=2)
    assert next_token_ids.equal(torch.tensor([[0, 2], [2, 1]]))


def test_multinomial():
    rng = torch.Generator()
    rng.manual_seed(239)

    logits = torch.tensor([[1.0, 4.0, 5.0]])
    next_token_ids = multinomial(logits, 1, rng)
    assert next_token_ids.equal(torch.tensor([[2]]))

    next_token_ids = multinomial(logits, 2, rng)
    assert next_token_ids.equal(torch.tensor([[2], [1]]))

    logits = torch.tensor([[10.0, 0.0, 9.0], [-math.inf, 4.0, 5.0]])
    next_token_ids = multinomial(logits, 1, rng)
    assert next_token_ids.equal(torch.tensor([[0, 2]]))

    next_token_ids = multinomial(logits, 2, rng)
    assert next_token_ids.equal(torch.tensor([[0, 1], [2, 2]]))


def test_vectorized_random_choice():
    rng = torch.Generator()
    rng.manual_seed(0)

    probs = torch.tensor([[1, 0, 0, 0]])
    sample = vectorized_random_choice(rng, probs)
    assert sample.shape == (1, 1)
    assert torch.equal(sample, torch.zeros((1, 1)))

    probs = torch.tensor([[1, 0, 0, 0]])
    sample = vectorized_random_choice(rng, probs, samples=3)
    assert sample.shape == (3, 1)
    assert torch.equal(sample, torch.zeros((3, 1)))

    probs = torch.tile(torch.tensor([[1, 0, 0, 0]]), (2, 1))
    sample = vectorized_random_choice(rng, probs)
    assert sample.shape == (1, 2)
    assert torch.equal(sample, torch.zeros((1, 2)))

    probs = torch.tensor([[1, 0, 0, 0], [0, 1, 0, 0]])
    sample = vectorized_random_choice(rng, probs, samples=3)
    assert sample.shape == (3, 2)
    assert torch.equal(sample, torch.tensor([[0, 1], [0, 1], [0, 1]]))

    probs = torch.tensor([[[1, 0, 0, 0], [0, 1, 0, 0]], [[0, 0, 1, 0], [0, 0, 0, 1]]])
    sample = vectorized_random_choice(rng, probs, samples=3)
    assert sample.shape == (3, 2, 2)
    assert torch.equal(
        sample, torch.tensor([[[0, 1], [2, 3]], [[0, 1], [2, 3]], [[0, 1], [2, 3]]])
    )



import math
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

import torch

from outlines.models import OpenAIAPI

if TYPE_CHECKING:
    from outlines.models.transformers import KVCacheType, Transformers
    from outlines.text.generate.sample import Sampler


class Sequence:
    """Represents a sequence generation method."""

    def __init__(
        self,
        model: "Transformers",
        max_tokens: Optional[int] = None,
        sampler: Optional["Sampler"] = None,
    ):
        """Create a `Sequence` instance.

        Parameters
        ----------
        model
            The instance of the model used to generate next-token probabilities.
        max_tokens
            The maximum number of tokens that will be generated if no termination
            condition is met.
        sampler
            The function used to draw samples.  Defaults to
            `outlines.text.generate.sample.multinomial`.  See
            `outlines.text.generate.sample.Sampler` for the expected form of
            such functions.

        """
        if isinstance(model, OpenAIAPI):
            raise TypeError("Cannot use guided generation with the OpenAI API.")

        self.model = model
        self.device = model.device
        self.max_tokens = max_tokens
        self.pad_token_id = torch.tensor(
            model.tokenizer.pad_token_id, device=model.device
        )
        if sampler is None:
            from outlines.text.generate.sample import multinomial

            self.sampler = multinomial
        else:
            self.sampler = sampler

    def create_proposal(
        self, generated_token_ids: torch.LongTensor, logits: torch.DoubleTensor
    ) -> torch.DoubleTensor:
        """Create a new proposal from the next-token logits."""
        return logits

    def is_finished(self, token_ids: torch.LongTensor) -> torch.BoolTensor:
        """Determine whether we should stop the generation."""
        raise NotImplementedError(
            "`Sequence.is_finished` must be implemented by subclasses."
        )

    def postprocess_completions(self, completions: List[str]) -> List[str]:
        return completions

    def step(
        self,
        rng: torch.Generator,
        num_prompt_tokens: int,
        token_ids: torch.LongTensor,
        attention_mask: torch.LongTensor,
        samples: int = 1,
        past_key_values: Optional["KVCacheType"] = None,
    ) -> Tuple[torch.LongTensor, torch.FloatTensor, Optional["KVCacheType"]]:
        """Generate one or several tokens that complete the input sequence.

        The sampling step consists in using a model to generate next-token
        logits and then sample `samples`-many new tokens from a categorical
        distribution parametrized by these logits.

        Parameters
        ----------
        rng
            Random number Generator instance.
        num_prompt_tokens
            The number of tokens in the prompt.
        token_ids
            The token sequences.  It has dimensions ``(n_seqs, n)`` for
            some sequence length ``n <= num_prompt_tokens``.
        samples
            The number of continuations to sample from the next-token probability
            distribution.

        Returns
        -------
        A tuple with an array of shape ``(samples, n_seqs, 1)``
        that contains the completed sequences (i.e. input token IDs and
        generated token IDs) and an array of shape
        ``(samples, n_seqs, vocab_size)`` that contains the next token
        probabilities.

        """
        probs, past_key_values = self.model.forward(
            token_ids, attention_mask, past_key_values
        )
        probs = self.create_proposal(token_ids[:, num_prompt_tokens:], probs)

        assert probs.shape[:-1] == token_ids.shape[:-1]

        next_token_ids = self.sampler(probs, samples, rng).unsqueeze(-1)

        probs = torch.broadcast_to(probs, (samples,) + probs.shape)

        return next_token_ids, probs, past_key_values

    def expand_attention_mask(
        self, attention_mask: torch.LongTensor
    ) -> torch.LongTensor:
        """Expand the attention mask after the last completion.

        Parameters
        ----------
        attention_mask
            An attention mask with shape ``(n_seqs, attention_mask_len)``.

        Returns
        -------
        A new attention mask with shape ``(n_seqs, attention_mask_len + 1)``.

        """
        attention_mask = torch.concatenate(
            [
                attention_mask,
                torch.ones(attention_mask.shape[:-1] + (1,), device=self.device),
            ],
            axis=-1,
        )
        return attention_mask

    @torch.inference_mode()
    def __call__(
        self,
        prompt: Union[str, List[str]],
        samples: int = 1,
        rng: Optional[torch.Generator] = None,
    ) -> Union[str, List[str]]:
        """Generate a new sequence given a prompt.

        Parameters
        ----------
        prompt
            The input prompt.
        samples
            The number of samples to generate for each prompt.

        Returns
        -------
        The full sequence that contains the prompts and the generated string.

        """

        token_ids, attention_mask = self.model.tokenizer.encode(prompt)

        token_ids = token_ids.squeeze(0)
        attention_mask = attention_mask.squeeze(0)

        token_ids = token_ids.to(self.device)
        attention_mask = attention_mask.to(self.device)

        if rng is None:
            rng = torch.Generator(device=self.device)
            rng.seed()

        orig_batch_shape = token_ids.shape[:-1]
        num_prompt_tokens = token_ids.shape[-1]

        token_ids = torch.broadcast_to(token_ids, (samples,) + token_ids.shape)
        attention_mask = torch.broadcast_to(
            attention_mask, (samples,) + attention_mask.shape
        )

        # We flatten the original batch and sample dimensions so that the
        # resulting shape we work in is simply `(num_of_sequences, tokens)`
        batch_size = samples * math.prod(orig_batch_shape)
        token_ids = token_ids.reshape((batch_size, num_prompt_tokens))
        attention_mask = attention_mask.reshape((batch_size, num_prompt_tokens))

        is_finished = torch.zeros(batch_size, dtype=torch.bool, device=self.device)
        unfinished_past_key_values = None

        while True:
            num_generated_tokens = token_ids.shape[-1] - num_prompt_tokens
            if torch.all(is_finished) or num_generated_tokens == self.max_tokens:
                break

            is_not_finished = ~is_finished

            # Draw samples only for the sequences that aren't finished
            unfinished_token_ids = token_ids[is_not_finished]
            unfinished_attention_mask = attention_mask[is_not_finished]
            unfinished_next_token_ids, _, past_key_values = self.step(
                rng,
                num_prompt_tokens,
                unfinished_token_ids,
                unfinished_attention_mask,
                past_key_values=unfinished_past_key_values,
            )
            unfinished_next_token_ids = unfinished_next_token_ids.squeeze(0)

            # Create an array for the next tokens of every sequence, including
            # the finished ones (but pad them)
            next_token_ids = torch.full(
                (batch_size, 1), self.pad_token_id, device=self.device
            )
            next_token_ids[is_not_finished] = unfinished_next_token_ids

            # TODO: Terminate if the sampled sequence is larger than the
            # context size of the model?
            token_ids = torch.concatenate([token_ids, next_token_ids], axis=-1)

            attention_mask = self.expand_attention_mask(attention_mask)

            local_is_finished = self.is_finished(
                token_ids[is_not_finished][:, num_prompt_tokens:]
            ).flatten()

            is_finished[is_not_finished] = local_is_finished

            if past_key_values:
                unfinished_past_key_values = tuple(
                    tuple(vv[~local_is_finished.to(vv.device)] for vv in v)
                    for v in past_key_values
                )

        result = self.model.tokenizer.decode(token_ids[:, num_prompt_tokens:])
        result = self.postprocess_completions(result)

        if len(result) == 1:
            return result[0]

        return result



import math

import interegular
import pytest
import torch

import outlines.text.generate as generate
from outlines.text.fsm import create_fsm_index_tokenizer, make_deterministic_fsm
from outlines.text.generate.regex import Regex


class Tokenizer:
    eos_token = "<EOS>"
    pad_token = None
    eos_token_id = 0
    pad_token_id = -1
    vocabulary = {"<EOS>": 0, "-": 1, "1": 2, "0.": 3, "431": 4, "a": 5, "A": 6}
    tokens = list(vocabulary.keys())
    special_tokens = {"<EOS>"}

    def encode(self, tokens):
        if not isinstance(tokens, (tuple, list)):
            tokens = [tokens]

        return [self.vocabulary[token] for token in tokens]

    def decode(self, token_ids):
        decoded = []
        for i in range(token_ids.shape[0]):
            decoded.append("".join([self.tokens[idx] for idx in token_ids[i]]))

        return decoded

    def convert_token_to_string(self, token):
        return token


class TokenizerWithEmpty(Tokenizer):
    vocabulary = {"<EOS>": 0, "-": 1, "1": 2, "0.": 3, "431": 4, "a": 5, "A": 6, "": 7}
    tokens = list(vocabulary.keys())


class Model:
    tokenizer = Tokenizer()
    device = "cpu"


class ModelWithEmpty:
    tokenizer = TokenizerWithEmpty()
    device = "cpu"


@pytest.mark.parametrize(
    "regex_string, valid_first_token, proposal",
    [
        (
            r"[A-Z]+",
            6,
            [-math.inf, -math.inf, -math.inf, -math.inf, -math.inf, -math.inf, 1.0],
        ),
        (
            r"[a-z]+",
            5,
            [-math.inf, -math.inf, -math.inf, -math.inf, -math.inf, 1.0, -math.inf],
        ),
        (
            r"(a|A)",
            6,
            [-math.inf, -math.inf, -math.inf, -math.inf, -math.inf, 1.0, 1.0],
        ),
        (r"\d+", 2, [-math.inf, -math.inf, 1.0, -math.inf, 1.0, -math.inf, -math.inf]),
        (r"\d+\.", 3, [-math.inf, -math.inf, 1.0, 1.0, 1.0, -math.inf, -math.inf]),
    ],
)
def test_regex_proposal(regex_string, valid_first_token, proposal):
    model = Model()
    generator = generate.regex(model, regex_string)

    logits = torch.ones(len(model.tokenizer.vocabulary))
    result = generator.create_proposal(torch.tensor([[]]), logits)
    assert torch.equal(result.squeeze(), torch.tensor(proposal))
    assert result.squeeze()[0] == -math.inf

    # The EOS token can be generated once the FSM is in an accept state
    result = generator.create_proposal(torch.tensor([[valid_first_token]]), logits)
    assert result.squeeze()[0] == 1


def test_regex_no_valid_transition():
    model = Model()
    with pytest.raises(ValueError, match="The vocabulary does not allow"):
        generate.regex(model, "aw")


@pytest.mark.parametrize(
    "input_ids, proposal",
    [
        ([[]], [[-math.inf, 1.0, 1.0, -math.inf, 1.0, -math.inf, -math.inf]]),
        ([[1]], [[-math.inf, -math.inf, 1.0, -math.inf, 1.0, -math.inf, -math.inf]]),
        ([[4]], [[1.0, -math.inf, 1.0, -math.inf, 1.0, -math.inf, -math.inf]]),
        (
            [[4], [2]],
            [
                [1.0, -math.inf, 1.0, -math.inf, 1.0, -math.inf, -math.inf],
                [1.0, -math.inf, 1.0, -math.inf, 1.0, -math.inf, -math.inf],
            ],
        ),
        (
            [[4, 0], [1, 2]],
            [
                [1.0, -math.inf, -math.inf, -math.inf, -math.inf, -math.inf, -math.inf],
                [1.0, -math.inf, 1.0, -math.inf, 1.0, -math.inf, -math.inf],
            ],
        ),
    ],
)
def test_integer_proposal(input_ids, proposal):
    model = Model()
    generator = generate.format(model, int)

    logits = torch.ones(len(model.tokenizer.vocabulary))
    result = generator.create_proposal(torch.tensor(input_ids), logits)
    assert torch.equal(
        result,
        torch.tensor(proposal),
    )


def test_choice_proposal():
    model = Model()
    generator = generate.choice(model, ["1", "431a", "431A-"])
    logits = torch.ones(len(model.tokenizer.vocabulary))
    result = generator.create_proposal(torch.tensor([[]]), logits)
    assert torch.equal(
        result,
        torch.tensor(
            [[-math.inf, -math.inf, 1.0, -math.inf, 1.0, -math.inf, -math.inf]]
        ),
    )

    result = generator.create_proposal(torch.tensor([[4]]), logits)
    assert torch.equal(
        result,
        torch.tensor(
            [[-math.inf, -math.inf, -math.inf, -math.inf, -math.inf, 1.0, 1.0]]
        ),
    )

    result = generator.create_proposal(torch.tensor([[4, 6]]), logits)
    assert torch.equal(
        result,
        torch.tensor(
            [[-math.inf, 1.0, -math.inf, -math.inf, -math.inf, -math.inf, -math.inf]]
        ),
    )


@pytest.mark.parametrize(
    "input_ids, proposal, with_empty",
    [
        ([[]], [[-math.inf, 1.0, 1.0, 1.0, 1.0, -math.inf, -math.inf]], False),
        (
            [[3]],
            [[-math.inf, -math.inf, 1.0, -math.inf, 1.0, -math.inf, -math.inf]],
            True,
        ),
    ],
)
def test_float_proposal(input_ids, proposal, with_empty):
    model = Model()
    generator = generate.format(model, float, allow_empty_tokens=with_empty)

    logits = torch.ones(len(model.tokenizer.vocabulary))
    result = generator.create_proposal(torch.tensor(input_ids), logits)
    assert torch.equal(
        result,
        torch.tensor(proposal),
    )


def test_Regex_precomputed():
    model = Model()
    choices = ["1", "431a", "431A-"]
    regex_str = r"(" + r"|".join(choices) + r")"

    regex_pattern = interegular.parse_pattern(regex_str)
    regex_fsm, _ = make_deterministic_fsm(regex_pattern.to_fsm().reduce())

    (
        states_to_token_maps,
        empty_token_ids,
    ) = create_fsm_index_tokenizer(regex_fsm, model.tokenizer)

    generator = Regex(
        model,
        regex_str,
        max_tokens=100,
        initial_state=regex_fsm.initial,
        final_states=regex_fsm.finals,
        states_to_token_maps=states_to_token_maps,
        empty_token_ids=empty_token_ids,
    )

    logits = torch.ones(len(model.tokenizer.vocabulary))
    result = generator.create_proposal(torch.tensor([[]]), logits)
    assert torch.equal(
        result,
        torch.tensor(
            [[-math.inf, -math.inf, 1.0, -math.inf, 1.0, -math.inf, -math.inf]]
        ),
    )
